; 
; This file contains a (incomplete) collection of MASM 6+
; macros, equates, and data and procedure definitions aimed
; at supporting real-mode DOS programs.
;
; The functions (are supposed to) preserve all registers
; that are not used to return a value.
; 

    ; DOS handles for standard input/output devices.
    HSTDIN   EQU 0
    HSTDOUT  EQU 1

    stdin     PROTO c :WORD,:WORD
    stdout    PROTO c :WORD
    szlen     PROTO c :WORD
    locate    PROTO c :BYTE,:BYTE
    w2dec     PROTO c :WORD,:WORD
    w2hex     PROTO c :WORD,:WORD
    dw2dec    PROTO c :DWORD,:WORD
    dw2hex    PROTO c :DWORD,:WORD
    b2bin     PROTO c :BYTE,:WORD
    szlower   PROTO c :WORD
    szupper   PROTO c :WORD      
    read_cmos PROTO c :BYTE

    print_far_string PROTO c :WORD,:WORD

; 

    ; This is the MASM32 reparg macro with *no* changes.
    ;
    ; "This macro replaces quoted text with a DATA section
    ; OFFSET and returns it in ADDR "name" format."
    ;
    reparg MACRO arg
      LOCAL nustr
      quot SUBSTR <arg>, 1, 1
      IFIDN quot,<">          ;; if 1st char = "
        .data
          nustr db arg, 0     ;; write arg to .DATA section
        .code
        EXITM <ADDR nustr>    ;; append name to ADDR operator
      ELSE
        EXITM <arg>           ;; else return arg
      ENDIF
    ENDM

    ; This is a VARARG version of the MASM32 print macro.
    ; It will accept either a quoted string, or the offset
    ; address of a null-terminated string in the data
    ; segment
    ;
    print MACRO varname:VARARG
      FOR arg, <varname>
        invoke stdout, reparg(arg)
      ENDM
    ENDM

    ; This is the MASM32 chr$ macro with *no* changes.
    ;
    ; "This macro takes quoted string and ascii numbers in
    ; the same format as a .DATA section entry. It returns
    ; the offset in the .DATA section of the data contained
    ; in its parameters."
    ;
    chr$ MACRO any_text:VARARG
      LOCAL txtname
      .data
        txtname db any_text,0
      .code
      EXITM <OFFSET txtname>
    ENDM

    ; This macro sets the cursor position for display
    ; page 0 to the specified row and column coordinates.
    ; The unusual feature is that either or both of the
    ; coordinate parameters may be left blank, causing
    ; the cursor position with respect to the blank
    ; coordinate to remain unchanged. For example,
    ; "loc 5" would position the cursor at row 5 without
    ; changing the column position, and "loc ,5" would
    ; position the cursor at column 5 without changing
    ; the row position.
    ;
    loc MACRO row,col
      IFNB <row>
        IFNB <col>
          invoke locate, row, col
        ELSE
          call cursorcol
          invoke locate, row, al
        ENDIF
      ELSE
        IFNB <col>
          call cursorrow
          invoke locate, al, col
        ELSE
          call cursorrow
          mov   ah, al
          call cursorcol
          invoke locate, ah, al
        ENDIF
      ENDIF
    ENDM

; 

.data
    inbuf       db 130 dup(0)
    outbuf      db 30 dup(0)
    crlf        db 13, 10, 0
    NL          dw offset crlf
    backspace   db 8, 0
    BS          dw OFFSET backspace
.code

; 
; This proc calls the Read File or Device function to read
; up to <blen> bytes from STDIN and store them in the
; specified buffer.
; 
stdin proc c uses bx cx dx npdest:WORD, blen:WORD
    mov   bx, HSTDIN
    mov   cx, blen
    mov   dx, npdest
    mov   ah, 3fh
    int   21h
    mov   bx, npdest
  @@:                         ; Replace CR with null
    cmp   BYTE PTR [bx], 0
    je    @F
    cmp   BYTE PTR [bx], 13
    je    @F
    inc   bx
    jmp   @B
  @@:
    mov   BYTE PTR [bx], 0
    ret
stdin endp

; 
; This proc calls the Write File or Device function to
; output the specified null-terminated string to STDOUT.
; 
stdout proc c uses ax bx cx dx npsz:WORD
    invoke szlen, npsz
    mov   cx, ax
    mov   bx, HSTDOUT
    mov   dx, npsz
    mov   ah, 40h
    int   21h
    ret
stdout endp

; 
; This proc returns the length of the specified
; null-terminated string in AX.
; 
szlen proc c uses bx npsz:WORD
    mov   bx, npsz
  runLoop:      
    cmp   BYTE PTR [bx], 0    ; Compare bytes until find null
    jz    @F
    inc   bx
    jmp   runLoop
  @@:      
    mov   ax, bx              ; Calc length and return in AX
    sub   ax, npsz
    ret
szlen endp

; 
; This proc calls the Flush Buffer, Read Keyboard Without
; Echo function and returns the character read in AL.
; 
waitkey proc c 
    mov   ax, 0c08h
    int   21h
    ret
waitkey endp

; 
; This proc calls the BIOS Set Cursor Position function
; to set the cursor position for display page 0 to
; <row>,<col> (upper left corner is 0,0).
; 
locate proc c uses ax bx dx row:BYTE, col:BYTE
    mov   ah, 2               ; Function number
    mov   bh, 0               ; Display page
    mov   dh, row
    mov   dl, col
    int   10h
    ret
locate endp

; 
; These procs return the row and column cursor positions
; for display page 0 in AL.
; 
cursorrow proc c uses es
    push  40h                 ; Set ES to the BIOS data area
    pop   es      
    mov   al, es:[51h]        ; Read the appropriate byte
    ret
cursorrow endp
cursorcol proc c uses es
    push  40h
    pop   es
    mov   al, es:[50h]
    ret
cursorcol endp

; 
; This proc clears the screen by calling the BIOS Scroll
; Current Page Up function with a scroll (row) count of zero,
; and sets the cursor position to the upper left corner of
; the screen. An 80x25 text mode screen is assumed.
; 
cls proc c uses ax bx cx dx
    mov   ah, 6               ; Function number
    mov   al, 0               ; Row count
    mov   bh, 7               ; Attribute for blanked lines
                              ; (white on black)
    mov   ch, 0               ; Top row of window
    mov   cl, 0               ; Left-most column of window
    mov   dh, 24              ; Bottom row of window
    mov   dl, 79              ; Right-most column of window
    int   10h
    invoke locate, 0, 0
    ret
cls endp

; 
; This proc converts a word to to a signed null-terminated
; decimal string and stores the result (6 characters max)
; in the specified buffer.
; 
w2dec proc c uses ax bx cx dx di src:WORD, npdest:WORD
    mov   ax, src
    mov   di, npdest
    test  ax, ax              ; Test for negative value
    jns   @F                  ; Make positive if negative
    neg   ax
    mov   BYTE PTR [di], '-'  ; Store sign in buffer
    inc   di                  ; Adjust DI for next character
  @@:      
    mov   bx, 10              ; Init divisor
    xor   cx, cx              ; Zero digit counter
  divLoop:
    xor   dx, dx              ; Zero MSW of dividend
    div   bx                  ; Divide value by 10
    push  dx                  ; Push remainder
    inc   cx                  ; Increment digit count
    or    ax, ax              ; Repeat if quotient > 0
    jnz   divLoop
  storeLoop:
    pop   ax                  ; Pop digit off stack
    add   al, '0'             ; Convert to ascii
    mov   [di], al            ; Store in buffer
    inc   di                  ; Adjust DI for next character
    loop  storeLoop           ; Repeat CX times
    mov   BYTE PTR [di], 0    ; Append trailing null
    ret
w2dec endp

; 
; This proc converts a word to to a null-terminated 4-character
; hex string and stores the result in the specified buffer.
; 
   .data
      hexChars db "0123456789ABCDEF"
   .code
w2hex proc c uses ax bx dx di src:WORD, npdest:WORD
    mov   dx, src
    mov   di, npdest
    add   di, 3               ; Set DI to end of buffer
    mov   BYTE PTR [di+1], 0  ; Append trailing null
  @@:
    mov   bx, dx              ; Copy current value into BX
    and   bx, 0Fh             ; Mask off all but lower nibble
    mov   al, hexChars[bx]    ; Load corresponding char into AL
    mov   [di], al            ; And store it in buffer
    shr   dx, 4               ; Shift next nibble into position
    dec   di                  ; Adjust DI for next character
    cmp   di, npdest          ; Repeat for all 4 characters
    jnb   @B
    ret
w2hex endp

; 
; This proc converts a dword to to an signed null-terminated
; decimal string and stores the result (11 characters max)
; in the specified buffer.
; 
dw2dec proc c uses ax bx cx dx di src:DWORD, npdest:WORD
    mov   eax, src
    mov   di, npdest
    test  eax, eax            ; Test for negative value
    jns   @F      
    neg   eax                 ; Make positive if negative
    mov   BYTE PTR[di], '-'   ; Store sign in buffer 
    inc   di
  @@:
    mov   ebx, 10             ; Init divisor
    xor   cx, cx              ; Zero digit counter
  divLoop:      
    xor   edx, edx            ; Zero MSDW of dividend
    div   ebx                 ; Divide value by 10
    push  dx                  ; Push remainder
    inc   cx                  ; Increment digit count
    or    eax, eax            ; Repeat if quotient > 0
    jnz   divLoop
  storeLoop:
    pop   ax                  ; Pop digit off stack
    add   al, '0'             ; Convert to ascii
    mov   [di], al            ; Store in buffer
    inc   di                  ; Adjust DI for next character
    loop  storeLoop           ; Repeat CX times
    mov   BYTE PTR [di], 0    ; Append trailing null
    ret
dw2dec endp

; 
; This proc converts a dword to to a null-terminated
; 8-character hex string and stores the result in the
; specified buffer.
;  
dw2hex proc c uses ax bx dx di src:DWORD, npdest:WORD
    mov   edx, src
    mov   di, npdest
    add   di, 7               ; Set DI to end of buffer
    mov   BYTE PTR [di+1], 0  ; Append trailing null
  @@:
    mov   ebx, edx            ; Copy current value into EBX
    and   ebx, 0Fh            ; Mask off all but lower nibble
    mov   al, hexChars[ebx]   ; Load corresponding char into AL
    mov   [di], al            ; And store it in buffer
    shr   edx, 4              ; Shift next nibble into position
    dec   di                  ; Adjust DI for next character
    cmp   di, npdest
    jnb   @B                  ; Repeat for all 8 characters
    ret
dw2hex endp

; 
; This proc converts a byte to to a null-terminated
; 8-character binary string and stores the result in
; the specified buffer.
; 
b2bin proc c uses ax bx cx dx src:BYTE, npdest:WORD
    mov   al, src
    mov   bx, npdest
    mov   cx, 8
    dec   bx                  ; DEC BX so first INC in loop
  shiftLoop:                   ; will not skip first bit
    inc   bx
    xor   dx, dx              ; Zero DX
    shl   al, 1               ; Shift next bit into carry flag
    adc   dl, '0'             ; Copy carry flag to DL and
                              ; convert to '0' or '1'
    mov   [bx], dl            ; Store in destination string
    dec   cx
    jnz   shiftLoop           ; Repeat for all bits
    mov   BYTE PTR [bx+1], 0  ; Append the terminating null
    ret
b2bin endp

; 
; This proc converts any upper-case characters in the
; specified null-terminated string to lower case.
; 
szlower proc c uses ax bx npsz:WORD
    mov   bx, npsz
  runLoop:
    mov   al, [bx]
    test  al, al              ; Test for terminating null
    jnz   @F
    ret                       ; Have reached end of string
  @@:      
    cmp   al, 'A'             ; Convert upper-case characters
    jb    @F                  ; only
    cmp   al, 'Z'             ;
    ja    @F                  ;
    add   al, 'a'-'A'         ;
    mov   [bx], al            ;
  @@:
    inc   bx
    jmp   runLoop
szlower endp

; 
; This proc converts any lower-case characters in the
; specified null-terminated string to upper case.
; 
szupper proc c uses ax bx npsz:WORD
    mov   bx, npsz
  runLoop:
    mov   al, [bx]
    test  al,al               ; Test for terminating null
    jnz   @F      
    ret                       ; Have reached end of string
  @@:
    cmp   al, 'a'             ; Convert lower-case characters
    jb    @F                  ; only
    cmp   al, 'z'             ;
    ja    @F                  ;
    sub   al, 'a'-'A'         ;
    mov   [bx], al            ;
  @@:
    inc   bx
    jmp   runLoop
szupper endp

; 
; This proc returns the value from the specified CMOS address
; in AL.
;
; Bit 7 of the CMOS address port is used to control the
; state of the NMI (Nonmaskable Interrupt). The remaining
; 7 bits are used to specify one of 128 possible addresses.
;
; Note that this procedure does not implement the logic
; necessary to reliably read the time and date registers.
;
; In practice, for anything resembling a recent system,
; the I/O delay is not necessary because the chipset
; inserts its own delay.
;
; This proc is essentially the READ_CMOS example from Frank
; van Gilluwe's book The Undocumented PC, First Edition,
; Addison-Wesley, 1994, ISBN 0-201-62277-7.
; 
IODELAY MACRO
    REPEAT 10
      jmp   $+2
    ENDM
ENDM
read_cmos proc c address:BYTE
    mov   al, address
    or    al, 80h             ; Set bit 7 of the address to
                              ; disable the NMI
    cli                       ; Disable hardware interrupts
    out   70h, al             ; Write to the CMOS address port
    IODELAY
    in    al, 71h             ; Read the addressed value THE
    IODELAY                   ; CMOS data port
    mov   ah, al              ; Preserve the value
    xor   al, al              ; Write a zero to the address
    out   70h, al             ; port to enable the NMI
    shr   ax, 8               ; Shift the value into AL
                              ; forcing AH to zero
    sti                       ; Enable hardware interrupts
    ret
read_cmos endp

; 
; This proc outputs the specified far string, enclosing it
; in single quotes.
; 
print_far_string proc c uses es bx str_seg:WORD, str_ptr:WORD
    LOCAL char[2]:BYTE

    print "'"
    mov   char[1], 0            ; Initialize null terminator
    mov   es, str_seg
    mov   bx, str_ptr
    .WHILE BYTE PTR es:[bx] != 0
      mov   al, es:[bx]
      mov   char, al
      print ADDR char
      inc   bx
    .ENDW
    print "'"
    ret
print_far_string endp
